import File from '../../file/File';
import Attr from '../../attribute/Attr';
import HTMLElement from '../html-element/HTMLElement';
import HTMLFormElement from '../html-form-element/HTMLFormElement';
import ValidityState from './ValidityState';
import HTMLInputElementSelectionModeEnum from './HTMLInputElementSelectionModeEnum';
/**
 * HTML Input Element.
 *
 * Reference:
 * https://developer.mozilla.org/en-US/docs/Web/API/HTMLInputElement
 *
 * Used as reference for some of the logic (like selection range):
 * https://github.com/jsdom/jsdom/blob/master/lib/jsdom/living/nodes/nodes/HTMLInputElement-impl.js (MIT licensed)
 */
export default class HTMLInputElement extends HTMLElement {
    formAction: string;
    formMethod: string;
    formNoValidate: boolean;
    name: string;
    _type: string;
    disabled: boolean;
    autofocus: boolean;
    required: boolean;
    _value: string;
    checked: boolean;
    defaultChecked: boolean;
    indeterminate: boolean;
    alt: string;
    height: number;
    src: string;
    width: number;
    files: File[];
    accept: string;
    allowdirs: string;
    autocomplete: string;
    min: string;
    max: string;
    minLength: number;
    maxLength: number;
    pattern: string;
    placeholder: string;
    readOnly: boolean;
    size: number;
    private _selectionStart;
    private _selectionEnd;
    private _selectionDirection;
    multiple: boolean;
    defaultValue: string;
    step: string;
    inputmode: string;
    /**
     * Returns value.
     *
     * @return Value.
     */
    get value(): string;
    /**
     * Sets value.
     *
     * @param value Value.
     */
    set value(value: string);
    /**
     * Sets type.
     *
     * @param type Type.
     */
    set type(type: string);
    /**
     * Returns type.
     *
     * @param type Type.
     */
    get type(): string;
    /**
     * Returns selection start.
     *
     * @return Selection start.
     */
    get selectionStart(): number;
    /**
     * Sets selection start.
     *
     * @param start Start.
     */
    set selectionStart(start: number);
    /**
     * Returns selection end.
     *
     * @return Selection end.
     */
    get selectionEnd(): number;
    /**
     * Sets selection end.
     *
     * @param end End.
     */
    set selectionEnd(end: number);
    /**
     * Returns selection direction.
     *
     * @return Selection direction.
     */
    get selectionDirection(): string;
    /**
     * Sets selection direction.
     *
     * @param direction Direction.
     */
    set selectionDirection(direction: string);
    /**
     * Returns the parent form element.
     *
     * @return Form.
     */
    get form(): HTMLFormElement;
    /**
     * Returns validity state.
     *
     * @returns Validity state.
     */
    get validity(): ValidityState;
    /**
     * Returns validity message.
     *
     * @return Validation message.
     */
    get validationMessage(): string;
    /**
     * Returns "true" if it will validate.
     *
     * @return "true" if it will validate.
     */
    get willValidate(): boolean;
    /**
     * Returns value as Date.
     *
     * @returns Date.
     */
    get valueAsDate(): Date;
    /**
     * Returns value as number.
     *
     * @return Number.
     */
    get valueAsNumber(): number;
    /**
     * Set selection range.
     *
     * @param start Start.
     * @param end End.
     * @param [direction="none"] Direction.
     */
    setSelectionRange(start: number, end: number, direction?: string): void;
    /**
     * Set range text.
     *
     * @param replacement Replacement.
     * @param [start] Start.
     * @param [end] End.
     * @param [direction] Direction.
     */
    setRangeText(replacement: any, start?: any, end?: any, selectionMode?: HTMLInputElementSelectionModeEnum): void;
    /**
     * Checks validity.
     *
     * @return "true" if validation does'nt fail.
     */
    checkValidity(): boolean;
    /**
     * Removes an Attr node.
     *
     * @override
     * @param attribute Attribute.
     */
    removeAttributeNode(attribute: Attr): void;
    /**
     * The setAttributeNode() method adds a new Attr node to the specified element.
     *
     * @override
     * @param attribute Attribute.
     * @returns Replaced attribute.
     */
    setAttributeNode(attribute: Attr): Attr;
    /**
     * Clones a node.
     *
     * @override
     * @param [deep=false] "true" to clone deep.
     * @return Cloned node.
     */
    cloneNode(deep?: boolean): HTMLInputElement;
    /**
     * Checks if private value is supported.
     *
     * @returns "true" if private value is supported.
     */
    private _isPrivateValueSupported;
    /**
     * Checks is selection is supported.
     *
     * @returns "true" if selection is supported.
     */
    private _isSelectionSupported;
}
